package cds;

import java.util.*;
import java.util.concurrent.*;


public class EratosthenesPlus {
    
    static final int NTHREADS = 1;
    
    public static void main(String[] args) throws InterruptedException {
        
        List<Integer> res = sieveParallelBasic(1000);
        //List<Integer> res = sieveSequential(3000);
        for(Integer p: res)
            System.out.println(p.toString());
        System.out.println("Total number of primes:"+Integer.toString(res.size()));
    }
    
    static List<Integer> sieveSequential(int upperLimit){
        
        if(upperLimit < 2) throw new IllegalArgumentException("The upper limit must be greather or equal than 2");
        
        List<Integer> primes = new LinkedList<>();
        boolean[] isPrime = new boolean[upperLimit+1];
        isPrime[0] = isPrime[1] = false;
        for(int i=2; i<=upperLimit; isPrime[i++]=true);
        
        System.out.println("initialized (sieveSequential)");
        
        int sqrtUL = (int)Math.floor(Math.sqrt(upperLimit));
        
        for(int i=2; i<=sqrtUL; i++) {
            if(isPrime[i]) {
                primes.add(i);
                
                int p = i*i;
                while (p<=upperLimit){
                    isPrime[p]=false;
                    p=p+i;
                }
            }
        }
        for (int i = sqrtUL+1; i<=upperLimit; i++)
            if(isPrime[i])
                primes.add(i);
        
        return primes;
    }
    
    static List<Integer> sieveParallelBasic(int upperLimit) throws InterruptedException{
        if(upperLimit<2) throw new IllegalArgumentException("The upper limit must be greather or equal than 2");
        
        List<Integer> primes = new LinkedList<>();
        java.util.concurrent.ExecutorService exec = Executors.newFixedThreadPool(NTHREADS);
        boolean[] isPrime = new boolean[upperLimit+1];
        isPrime[0] = isPrime[1] = false;
        for(int i=2; i<=upperLimit; isPrime[i++]=true);
        
        List<Future<Integer>> futureResults = new ArrayList<>();
        
        System.out.println("initialized (sieveParallelBasic)");
        
        int sqrtUL = (int)Math.floor(Math.sqrt(upperLimit));
        
        for(int i=2; i<=sqrtUL; i++) {
            if(isPrime[i]) {
                primes.add(i);
                int p = i*i;
                //the interval [p,upperlimit] is evenly subdivided among tasks,
                //and in each iteration a sub-interval is computed 
                //and the related task submitted
                int intervalInit=p;
                int intervalEnd;
                for(int t=0; t<NTHREADS;t++){
                    if (t!=NTHREADS-1)
                        intervalEnd = intervalInit+((upperLimit-intervalInit)/(NTHREADS-t));   
                    else 
                        intervalEnd = upperLimit;
                    //submit with intervalInit, IntervalEnd
                    Callable task = new SieverBasic(isPrime, intervalInit, intervalEnd, i);
                    futureResults.add(exec.submit(task));
                    
                    intervalInit = intervalEnd+1;    
                }
                
                //now we wait for threads to complete 
                //(this is the only goal for Futures here)
                for(Future<Integer> f: futureResults) {
                    try {
                        f.get();
                    } catch (InterruptedException | ExecutionException ex) {}
                }
                futureResults.clear();
            }
        }
        exec.shutdown();
        
        for (int i = sqrtUL+1; i<=upperLimit; i++)
            if(isPrime[i])
                primes.add(i);
        
        return (List<Integer>)primes;
    }
    
    
}

class SieverBasic implements Callable<Integer>{
    
    private final int from, to, primeStep;
    private boolean[] target;
    
    public SieverBasic(boolean[] target, int from, int to, int primeStep){
        int rest=from%primeStep;
        this.from = rest==0 ? from : from+(primeStep-rest);
        this.to = to;
        this.primeStep = primeStep;
        this.target = target;
    }

    @Override
    public Integer call() {
        for(int i=from; i<=to; i+=primeStep)
            target[i] = false;
        return 0;
    }
    
}
