package cds;

import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * Exercise to experiment a tread starvation LIVELOCK.
 * Same setting as ThreadStarvationDeadlockTest, but in this case busy waiting 
 * is used.
 * 
 * @author alessio
 */

public class ThreadStarvationLivelockTest {
    static final int NUMTASKS=8;
    static final int POOLSIZE=5;
    static final int INITVALUE=8; //initial number for the shared counter
    private final ExecutorService myExecutor;
    private final IncClassBusy[] myTasks;
    
    public ThreadStarvationLivelockTest(ExecutorService exc) {
        myExecutor = exc;       
        // all the tasks are created
        myTasks = new IncClassBusy[NUMTASKS];
        for (int i=0; i<myTasks.length; myTasks[i++]=new IncClassBusy());  
    }
    
    private void runAll(){
        for(Runnable t: myTasks)
            myExecutor.execute(t);
        myExecutor.shutdown();
    }
    
    public static void main(String[] arg){
        ThreadStarvationLivelockTest myInst=
            new ThreadStarvationLivelockTest(Executors.newFixedThreadPool(POOLSIZE));
        
        myInst.runAll();
    }



    static class IncClassBusy implements Runnable{
        
        private static final AtomicInteger c = new AtomicInteger(INITVALUE);

        public void IncClassBusy(){        
        }

        @Override
        public void run() {
            int taskID = INITVALUE - c.getAndDecrement();
            
            //start message
            String msg = "";
            for (int i=0; i++<taskID; msg+="  ");
            msg += "Task #"+Integer.toString(taskID)+": STARTED!";
            System.out.println(msg);
            
            while(c.get()>0) {} // busy waiting
            
            //end message
            msg = "";
            for (int i=0; i++<taskID; msg+="  ");
            msg += "Task #"+Integer.toString(taskID)+": DONE!";
            System.out.println(msg);
        }
    }
}
